"""Represents a disy Cadenza analytics extension and holds its configuration. In conenction with the
`cadenza anayltics extension service` the extension handels the processing of analytics requests, when
invoked via HTTP POST on the relative path."""
import json
from io import StringIO
from typing import Callable, List

import pandas as pd
from flask import Response, request

from cadenzaanalytics.data.analytics_extension import AnalyticsExtension
from cadenzaanalytics.data.attribute_group import AttributeGroup
from cadenzaanalytics.data.extension_type import ExtensionType
from cadenzaanalytics.data.parameter import Parameter
from cadenzaanalytics.request.analytics_request import AnalyticsRequest
from cadenzaanalytics.request.request_metadata import RequestMetadata
from cadenzaanalytics.response.extension_response import ExtensionResponse


class CadenzaAnalyticsExtension:
    """Class representing a Cadenza analytics extension.
    """
    def __init__(self,
                 relative_path: str,
                 analytics_function: Callable[[RequestMetadata, pd.DataFrame], ExtensionResponse],
                 print_name: str,
                 extension_type: ExtensionType,
                 attribute_groups: List[AttributeGroup],
                 parameters: List[Parameter] = None):
        self._relative_path = relative_path
        self._analytics_function = analytics_function

        self._analytics_extension = AnalyticsExtension(print_name, extension_type, attribute_groups, parameters)

    @property
    def relative_path(self) -> str:
        """Getter for the relative path of the extension.

        Returns
        -------
        str
            The relative path of the extension.
        """
        return self._relative_path

    @property
    def print_name(self) -> str:
        """Getter for the printable name of the extension.

        Returns
        -------
        str
            The printable name of the extension.
        """
        return self._analytics_extension.print_name

    @property
    def extension_type(self) -> ExtensionType:
        """Getter for the type of the extension.

        Returns
        -------
        str
            The type of the extension.
        """
        return self._analytics_extension.extension_type

    def handle_request(self) -> Response:
        """Handle the processing of extension requests.

        Returns
        -------
        Response
            The response to the request.
        """
        analytics_request = self._get_request_data(request)

        analytics_response = self._analytics_function(analytics_request.metadata, analytics_request.data)

        return analytics_response.get_response(analytics_request.metadata.get_columns(), analytics_request.data)

    def get_capabilities(self) -> Response:
        """Get the capabilities of the extension.

        Returns
        -------
        Response
            The capabilities of the extension.
        """
        return Response(response=self._analytics_extension.to_json(), status=200, mimetype="application/json")

    def _get_request_data(self, multipart_request) -> AnalyticsRequest:
        metadata_dict = json.loads(multipart_request.form['metadata'])
        metadata = RequestMetadata(metadata_dict)

        if metadata.has_columns():
            type_mapping = {}
            for column in metadata.get_columns():
                type_mapping[column.name] = column.data_type.pandas_type()
            csv_data = StringIO(multipart_request.form['data'])
            df_data = pd.read_csv(csv_data, sep=";", dtype=type_mapping)
        else:
            df_data = pd.DataFrame()

        return AnalyticsRequest(metadata, df_data)



from enum import Enum


class MeasureAggregation(Enum):
    """A class representing various aggregation functions for measures such as count, countDistinct,
    sum, average, min and max.
    """
    COUNT = "count"
    COUNT_DISTINCT = "countDistinct"
    SUM = "sum"
    AVERAGE = "average"
    MIN = "min"
    MAX = "max"

    def __str__(self):
        return self.value



import csv
import sys
from typing import List

from pandas import DataFrame

from cadenzaanalytics.data.column_metadata import ColumnMetadata
from cadenzaanalytics.response.extension_data_response import ExtensionDataResponse


class CsvResponse(ExtensionDataResponse):
    """A class representing a CSV response from an extension.

    Parameters
    ----------
    ExtensionDataResponse : type
        The base extension data response type from which CsvResponse inherits.
    """
    def __init__(self, data: DataFrame, column_metadata: List[ColumnMetadata]):
        content_type = 'text/csv'
        super().__init__(content_type)

        self._data = data
        self._column_meta_data = column_metadata

    def get_response(self, original_column_metadata: List[ColumnMetadata], original_data: DataFrame):
        """Get the CSV response.

        Returns
        -------
        Response
            The CSV response.
        """
        python_3_12 = (3, 12)
        if sys.version_info >= python_3_12:
            csv_data = self._data.to_csv(
                sep=';',
                encoding='utf-8',
                quoting=csv.QUOTE_NOTNULL,
                index=False,
                na_rep=None,  # missing/None/Null values are sent without quotes
                quotechar='"',
                lineterminator='\r\n')
        else:
            csv_data = self._data.to_csv(
                sep=';',
                encoding='utf-8',
                quoting=csv.QUOTE_ALL,
                index=False,
                quotechar='"',
                lineterminator='\r\n')
            # Needed to make sure we sent NULL/None values and not empty strings.
            # This replacement might be problematic for escaped quotes within a string.
            # As the proper solution is to use python 3.12,
            # we do not support more sophisticated support for missing values here.
            csv_data = csv_data.replace('""', '')

        return self._create_response(csv_data, self._column_meta_data)



from enum import Enum


class ParameterType(Enum):
    """A class representing parameter types.

    Returns
    -------
    str
        A string representing the parameter type.
    """
    STRING = "string"
    INT64 = "int64"
    FLOAT64 = "float64"
    ZONEDDATETIME = "zonedDateTime"
    GEOMETRY = "geometry"
    SELECT = "select"
    BOOLEAN = "boolean"

    def __str__(self):
        return self.value



from typing import List, Dict, Optional

from cadenzaanalytics.data.column_metadata import ColumnMetadata


# pylint: disable=protected-access
class RequestMetadata:
    """A class representing the metadata for an analytics request.
    """
    def __init__(self, request_metadata: dict):
        self._request_metadata = request_metadata

    def get_column(self, name: str) -> Optional[ColumnMetadata]:
        """Returns the column metadata object for a specific column accessed by its name.

        Parameters
        ----------
        name : str
            The name of the column.

        Returns
        -------
        ColumnMetadata | None
            Metadata for the column if found, else None.
        """
        if self.has_columns():
            for column in self._get_columns():
                if column["name"] == name:
                    return ColumnMetadata._from_dict(column)

        return None

    def get_first_column_of_attribute_group(self, attribute_group) -> Optional[ColumnMetadata]:
        """Returns the first column metadata object of the given attribute group.

        Returns
        -------
        Optional[ColumnMetadata]
            The column metadata of the first column in the given attribute group. If no column metadata for the
            attribute group was send None is returned.
        """
        if self.has_columns():
            for column in self._get_columns():
                if column['attributeGroupName'] == attribute_group:
                    return ColumnMetadata._from_dict(column)
        return None

    def get_columns_by_attribute_group(self) -> Dict[str, List[ColumnMetadata]]:
        """Returns all column metadata objects grouped by its attribute groups.

        Returns
        -------
        Dict[str, List[ColumnMetadata]]
            A dictionary where the keys are the attribute group names and values are lists of corresponding
             column metadata objects.
        """
        grouped_columns = {}
        columns = self._get_columns() if self.has_columns() else []
        for column in columns:
            if column['attributeGroupName'] not in grouped_columns:
                grouped_columns[column['attributeGroupName']] = []

            grouped_columns[column['attributeGroupName']].append(ColumnMetadata._from_dict(column))

        return grouped_columns

    def get_columns(self) -> List[ColumnMetadata]:
        """Returns a list of all column metadata objects.

        Returns
        -------
        List[ColumnMetadata]
            A list of all column metadata objects.
        """
        columns = self._get_columns() if self.has_columns() else []
        return [ColumnMetadata._from_dict(column) for column in columns]

    def get_parameters(self) -> Dict[str, str]:
        """Returns all parameters of the request.

        Returns
        -------
        dict[str, str]
            Parameters of the request.
        """
        return self._request_metadata['parameters']

    def get_parameter(self, name: str) -> Optional[str]:
        """Returns a specific parameter value.

        Parameters
        ----------
        name : str
            The name of the parameter.

        Returns
        -------
        str
            The value of the parameter if found, else an empty string.
        """
        if name in self._request_metadata['parameters']:
            return self._request_metadata['parameters'][name]
        return None

    def has_columns(self) -> bool:
        """Check if the analytics request has columns with corresponding metadata.

        Returns
        -------
        bool
            True if the request has columns with corresponding metadata, False otherwise.
        """
        return (len(self._request_metadata['dataContainers']) > 0
                and "columns" in self._request_metadata['dataContainers'][0])

    def _get_columns(self):
        return self._request_metadata['dataContainers'][0]['columns']



from enum import Enum


class ExtensionType(Enum):
    """A class representing the different types of extensions such as visualization, enrichment, calculation.
    """
    VISUALIZATION = "visualization"
    ENRICHMENT = "enrichment"
    CALCULATION = "calculation"

    def __str__(self):
        return self.value



from cadenzaanalytics.request.request_metadata import RequestMetadata


class AnalyticsRequest:
    """A class representing an analytics request.
    """
    def __init__(self, metadata: RequestMetadata, data):
        self._metadata = metadata
        self._data = data

    @property
    def metadata(self) -> RequestMetadata:
        """Get the metadata associated with the request.

        Returns
        -------
        RequestMetadata
            The metadata associated with the request.
        """
        return self._metadata

    @property
    def data(self):
        """Get the data payload of the request.

        Returns
        -------
        object
            The data associated with the request.
        """
        return self._data
